#!/usr/bin/python
import re

# Composite codes we are interested in (HIM cannot use any other)
interested = [ 0x0300, 0x0301, 0x0302, 0x0303, 0x0304, 0x032e, 0x0307, 0x0308,
	0x030a, 0x030b, 0x030c, 0x0327, 0x0328, 0x0323, 0x0309, 0x031b]
interested.sort()

# Auxiliar function for parsing the decomposition segment in UnicodeData
hex_p = re.compile(r"\s+")
def parse_decomp(str):
	m = hex_p.split(str)
	f = filter(lambda x: x[0] != '<', m)
	return [int(x, 16) for x in f]

# First read all composition exclusions
p = re.compile(r"(?P<value>[0-9A-F]+)")
f = open('CompositionExclusions.txt', 'r')
exclusions = set()

for line in f:
	m = p.match(line)
	if m is None:
		continue
	
	value = int(m.group('value'), 16)
	
	exclusions.add(value)

f.close()

# Now read decomposition data
p = re.compile(r"(?P<value>[0-9A-F]+);(?P<name>[^;]*);[^;]*;(?P<canonical_class>\d+);[^;]*;(?P<decomp>[^;]*);")
f = open('UnicodeData.txt', 'r')
composite = {}

for line in f:
	m = p.match(line)
	if m is None:
		print "#warning invalid line:", line
		continue
	
	value = int(m.group('value'), 16)
	name = m.group('name')
	segment = m.group('decomp')
	if len(segment) > 0:	
		compat = segment[0] == '<'
		if not compat and value not in exclusions:
			decomp = parse_decomp(segment)
			first = 0
			second = decomp[0]
			if len(decomp) > 1:
				first = second
				second = decomp[1]
			
			if second not in composite:
				composite[second] = {first: value}
			else:
				composite[second][first] = value

# Nothing left to be collected.  Let there be light!
print """
/* This file is autogenerated by builder.py from UnicodeData.txt */
/* Do not edit; instead edit builder.py and regenerate. */

struct item {
	unsigned short first;
	unsigned short result;
};

struct table {
	unsigned short second;
	unsigned short data_size;
	const struct item * data;
};

"""

for second in interested:
	if second not in composite:
		print "#warning data for composite 0x%04x not found" % second
		continue
	
	print "static const struct item values_for_%04x[] = {" % second
	
	items = composite[second].items()
	items.sort()
	for first, value in items:
		print "\t{0x%x, 0x%x}," % (first, value)
	
	print "};"
	print ""

print ""

print "static const struct table composite_table[] = {"
for second in interested:
	if second not in composite:
		continue
	
	data_size = len(composite[second])
	print "\t{0x%x, %u, values_for_%04x}, " % (second, data_size, second)
print "};"

f.close()
